module GENA;

export {
	redef enum Log::ID += { GENA_LOG };

    ## Default logging policy hook for GENA_LOG.
    global log_policy: Log::PolicyHook;

	type Info: record {
		## Time
		ts: time &log &default=network_time();

		## Unique ID for the connection
		uid: string &log;

		## The connection's 4-tuple of endpoint addresses/ports
		id: conn_id &log;

		## Is orig side of connection
		is_orig: bool &log;

		## The protcol.
		proto: transport_proto &optional &log;

		## The operation
		operation: string &log &optional;

        ## The publication path
        pub_path: string &log &optional;

        ## The status code
        status_code: string &log &optional;

        ## The HTTP version
        http_ver: string &log &optional;

		## The host
		host: string &log &optional;

		## The USN
		usn: string &log &optional;

		## The target (notification or search)
		target: string &log &optional;

		## The target sub type
		target_sub: string &log &optional;

		## The subscription ID
		sid: string &log &optional;

		## The sequence
		seq: string &log &optional;

		## The callback
		callback: string &log &optional;

		## The user agent
		user_agent: string &log &optional;

		## The server info
		server: string &log &optional;

		## Content length
		content_len: count &log &optional;

		## Transfer encoding
		transfer_encoding: string &log &optional;

		## Remaining header names
		remaining_header_names: vector of string &log &optional;

		## Remaining header values
		remaining_header_values: vector of string &log &optional;

        ## The Payload
        payload: string &log &optional;
	};

	# Set to true if you want a gena.log.  Caution!!! GENA is chatty and could create a lot of logs!
	option enable_logging: bool = T;

	### Events ###

	# Raised for every GENA message
	global GENA::message: event(c: connection, is_orig: bool, operation: string, pub_path: string, 
								status_code: string, http_ver: string, payload: string);

	# Raised for every GENA header
	global GENA::header: event(c: connection, is_orig: bool, name: string, value: string);

	# Log event
	global GENA::log_gena: event(rec: GENA::Info);
}

redef record connection += {
	gena: Info &optional;
};

function new_info(c: connection)
	{
	c$gena = [$uid=c$uid, $id=c$id, $is_orig=T];
	c$gena$remaining_header_names = vector();
	c$gena$remaining_header_values = vector();
	}

function set_session(c: connection)
	{
	if ( ! c?$gena )
		{
		new_info(c);
		}
	}

event GENA::message(c: connection, is_orig: bool, operation: string, pub_path: string, 
					status_code: string, http_ver: string, payload: string)
	{
	if ( ! c?$gena )
		return;
	
	c$gena$is_orig = is_orig;
    c$gena$proto = get_port_transport_proto(c$id$resp_p);
	c$gena$operation = operation;
	if (|pub_path| > 0)
	    c$gena$pub_path = pub_path;
	if (|status_code| > 0)
		c$gena$status_code = status_code;
    c$gena$http_ver = http_ver;
    c$gena$payload = payload;

	if ( enable_logging )
		Log::write(GENA::GENA_LOG, c$gena);

	# Clear out data for our next message
	new_info(c);
	}

event GENA::header(c: connection, is_orig: bool, name: string, value: string)
	{
	set_session(c);

	local lil_name = to_lower(name);

	switch(lil_name) 
		{
		case "user-agent":
			c$gena$user_agent = value;
			break;
		case "nt":
			fallthrough;
		case "st":
			c$gena$target = value;
			break;
		case "nts":
			c$gena$target_sub = value;
			break;
		case "callback":
			c$gena$callback = value;
			break;
		case "usn":
			c$gena$usn = value;
			break;
		case "server":
			c$gena$server = value;
			break;
		case "sid":
			c$gena$sid = value;
			break;
		case "seq":
			c$gena$seq = value;
			break;
		case "host":
			c$gena$host = value;
			break;
		case "content-length":
			c$gena$content_len = to_count(value);
			break;
		case "transfer-encoding":
			c$gena$transfer_encoding = value;
			break;
		default:
			c$gena$remaining_header_names += name;
			c$gena$remaining_header_values += value;
			break;
		}
	}

event zeek_init() &priority=5 
	{
	Log::create_stream(GENA::GENA_LOG, [$columns=Info, $ev=log_gena, $path="gena", $policy=log_policy]);
	}
